# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.server` module`
"""

from unittest import TestCase
import ssl

from .helpers import DummySocket
from degu.base import _TYPE_ERROR
from degu.sslhelpers import random_id
from degu.misc import TempPKI
from degu import base, client


# Good type and value permutations for the CLient *address*:
GOOD_ADDRESSES = (

    # 2-tuple for AF_INET or AF_INET6:
    ('www.wikipedia.org', 80),
    ('208.80.154.224', 80),
    ('2620:0:861:ed1a::1', 80),

    # 4-tuple for AF_INET6:
    ('2620:0:861:ed1a::1', 80, 0, 0),
    ('fe80::e8b:fdff:fe75:402c', 80, 0, 3),  # Link-local

    # str for AF_UNIX:
    '/tmp/my.socket',

    # bytes for AF_UNIX (Linux abstract name):
    b'\x0000022',
)

# Some bad address permutations:
BAD_TUPLE_ADDRESSES = (
    ('::1',),
    ('127.0.0.1',),
    ('::1', 5678, 0),
    ('127.0.0.1', 5678, 0),
    ('::1', 5678, 0, 0, 0),
    ('127.0.0.1', 5678, 0, 0, 0),
)


class TestFunctions(TestCase):
    def test_build_client_sslctx(self):
        # Bad sslconfig type:
        with self.assertRaises(TypeError) as cm:
            client.build_client_sslctx('bad')
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('sslconfig', dict, str, 'bad')
        )
 
        # The remaining test both build_client_sslctx() directly, and the
        # pass-through from _validate_client_sslctx():
        client_sslctx_funcs = (
            client.build_client_sslctx,
            client._validate_client_sslctx,
        )

        # Bad sslconfig['check_hostname'] type:
        for func in client_sslctx_funcs:
            with self.assertRaises(TypeError) as cm:
                func({'check_hostname': 0})
            self.assertEqual(str(cm.exception),
                _TYPE_ERROR.format("sslconfig['check_hostname']", bool, int, 0)
            )

        # sslconfig['key_file'] without sslconfig['cert_file']:
        for func in client_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func({'key_file': '/my/client.key'})
            self.assertEqual(str(cm.exception), 
                "sslconfig['key_file'] provided without sslconfig['cert_file']"
            )

        # Non absulute, non normalized paths:
        good = {
            'ca_file': '/my/sever.ca',
            'ca_path': '/my/sever.ca.dir',
            'cert_file': '/my/client.cert',
            'key_file': '/my/client.key',
        }
        for key in good.keys():
            # Relative path:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = 'relative/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

            # Non-normalized path with directory traversal:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = '/my/../secret/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

            # Non-normalized path with trailing slash:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = '/sorry/very/strict/'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

        # Empty sslconfig, will verify against system-wide CAs, and
        # check_hostname should default to True:
        for func in client_sslctx_funcs:
            sslctx = func({})
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # We don't not allow check_hostname to be False when verifying against
        # the system-wide CAs:
        for func in client_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func({'check_hostname': False})
            self.assertEqual(str(cm.exception),
                'check_hostname must be True when using default verify paths'
            )

        # Should work fine when explicitly providing {'check_hostname': True}:
        for func in client_sslctx_funcs:
            sslctx = func({'check_hostname': True})
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # Authenticated client sslconfig:
        pki = TempPKI()
        sslconfig = pki.client_sslconfig
        self.assertEqual(set(sslconfig),
            {'ca_file', 'cert_file', 'key_file', 'check_hostname'}
        )
        self.assertIs(sslconfig['check_hostname'], False)
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, False)

        # check_hostname should default to True:
        del sslconfig['check_hostname']
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # Anonymous client sslconfig:
        sslconfig = pki.anonymous_client_sslconfig
        self.assertEqual(set(sslconfig), {'ca_file', 'check_hostname'})
        self.assertIs(sslconfig['check_hostname'], False)
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, False)

        # check_hostname should default to True:
        del sslconfig['check_hostname']
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)


class TestClient(TestCase):
    def test_init(self):
        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            client.Client(1234)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        for address in BAD_TUPLE_ADDRESSES:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                client.Client(address)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # Non-absolute/non-normalized AF_UNIX filename:
        with self.assertRaises(ValueError) as cm:
            client.Client('foo')
        self.assertEqual(str(cm.exception),
            "address: bad socket filename: 'foo'"
        )

        # Good address type and value permutations:
        for address in GOOD_ADDRESSES:
            inst = client.Client(address)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option:
            my_host = '.'.join([random_id() for i in range(3)])
            inst = client.Client(address, host=my_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': my_host})
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding `host` option with `None`:
            inst = client.Client(address, host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': None})
            self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `timeout` option:
            inst = client.Client(address, timeout=17)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'timeout': 17})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 17)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `on_connect` option:
            inst = client.Client(address, on_connect=None)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            with self.assertRaises(TypeError) as cm:
                client.Client(address, on_connect='hello')
            self.assertEqual(str(cm.exception),
                "on_connect: not callable: 'hello'"
            )

            def my_on_connect(conn):
                return True
            inst = client.Client(address, on_connect=my_on_connect)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 65)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test overriding all the options together:
            options = {
                'host': my_host,
                'timeout': 16.9,
                'on_connect': my_on_connect,
            }
            inst = client.Client(address, **options)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, options)
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 16.9)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test unsupported options:
            with self.assertRaises(TypeError) as cm:
                client.Client(address, foo=17)
            self.assertEqual(str(cm.exception),
                'unsupported Client() **options: foo'
            )
            with self.assertRaises(TypeError) as cm:
                client.Client(address, foo=17, bar=18)
            self.assertEqual(str(cm.exception),
                'unsupported Client() **options: bar, foo'
            )

    def test_repr(self):
        class Custom(client.Client):
            pass

        for address in GOOD_ADDRESSES:
            inst = client.Client(address)
            self.assertEqual(repr(inst), 'Client({!r})'.format(address))
            inst = Custom(address)
            self.assertEqual(repr(inst), 'Custom({!r})'.format(address))

    def test_connect(self):
        class ClientSubclass(client.Client):
            def __init__(self, sock, host, on_connect=None):
                self.__sock = sock
                self._base_headers = {'host': host}
                self.on_connect = on_connect

            def create_socket(self):
                return self.__sock

        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host)
        self.assertIsNone(inst.on_connect)
        conn = inst.connect()
        self.assertIsInstance(conn, base.Connection)
        self.assertIs(conn.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertEqual(sock._calls, [])

        # Should return a new Connection instance each time:
        conn2 = inst.connect()
        self.assertIsNot(conn2, conn)
        self.assertIsInstance(conn2, base.Connection)
        self.assertIs(conn2.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertEqual(sock._calls, [])

        # on_connect() returns True:
        def on_connect_true(conn):
            return True
        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host, on_connect_true)
        self.assertIs(inst.on_connect, on_connect_true)
        conn = inst.connect()
        self.assertIsInstance(conn, base.Connection)
        self.assertIs(conn.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertEqual(sock._calls, [])

        # on_connect() does not return True:
        def on_connect_false(conn):
            return 1
        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host, on_connect_false)
        self.assertIs(inst.on_connect, on_connect_false)
        with self.assertRaises(ValueError) as cm:
            conn = inst.connect()
        self.assertEqual(str(cm.exception), 'on_connect() did not return True')


class TestSSLClient(TestCase):
    def test_init(self):
        # sslctx is not an ssl.SSLContext:
        with self.assertRaises(TypeError) as cm:
            client.SSLClient('foo', None)
        self.assertEqual(str(cm.exception), 'sslctx must be an ssl.SSLContext')

        # Bad SSL protocol version:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.protocol must be ssl.PROTOCOL_TLSv1_2'
        )

        # Note: Python 3.3.4 (and presumably 3.4.0) now disables SSLv2 by
        # default (which is good); Degu enforces this (also good), but because
        # we cannot unset the ssl.OP_NO_SSLv2 bit, we can't unit test to check
        # that Degu enforces this, so for now, we set the bit here so it works
        # with Python 3.3.3 still; see: http://bugs.python.org/issue20207
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        sslctx.options |= ssl.OP_NO_SSLv2

        # not (options & ssl.OP_NO_COMPRESSION)
        sslctx.options |= ssl.OP_NO_SSLv2
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_NO_COMPRESSION'
        )

        # verify_mode is not ssl.CERT_REQUIRED:
        sslctx.options |= ssl.OP_NO_COMPRESSION
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.verify_mode must be ssl.CERT_REQUIRED'
        )

        #############################
        # Good sslctx from here on...
        sslctx.verify_mode = ssl.CERT_REQUIRED

        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            client.SSLClient(sslctx, 1234)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        for address in BAD_TUPLE_ADDRESSES:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                client.SSLClient(sslctx, address)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # Non-absolute/non-normalized AF_UNIX filename:
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, 'foo')
        self.assertEqual(str(cm.exception),
            "address: bad socket filename: 'foo'"
        )

        # Good address type and value permutations:
        for address in GOOD_ADDRESSES:
            inst = client.SSLClient(sslctx, address)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option:
            my_host = '.'.join([random_id() for i in range(3)])
            inst = client.SSLClient(sslctx, address, host=my_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': my_host})
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option with `None`:
            inst = client.SSLClient(sslctx, address, host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': None})
            self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `ssl_host` options:
            my_ssl_host = '.'.join([random_id(10) for i in range(4)])
            inst = client.SSLClient(sslctx, address, ssl_host=my_ssl_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'ssl_host': my_ssl_host})
            self.assertIs(inst.ssl_host, my_ssl_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `ssl_host` option with `None`:
            inst = client.SSLClient(sslctx, address, ssl_host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'ssl_host': None})
            self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `timeout` option:
            inst = client.SSLClient(sslctx, address, timeout=17)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'timeout': 17})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 17)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `on_connect` option:
            with self.assertRaises(TypeError) as cm:
                client.SSLClient(sslctx, address, on_connect='hello')
            self.assertEqual(str(cm.exception),
                "on_connect: not callable: 'hello'"
            )
            def my_on_connect(conn):
                return True
            inst = client.SSLClient(sslctx, address, on_connect=my_on_connect)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test overriding the `on_connect` option with `None`:
            inst = client.SSLClient(sslctx, address, on_connect=None)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 65)
            self.assertIsNone(inst.on_connect)

            # Test overriding all the options together:
            options = {
                'host': my_host,
                'ssl_host': my_ssl_host,
                'timeout': 16.9,
                'on_connect': my_on_connect,
            }
            inst = client.SSLClient(sslctx, address, **options)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, options)
            self.assertIs(inst.host, my_host)
            self.assertIs(inst.ssl_host, my_ssl_host)
            self.assertEqual(inst.timeout, 16.9)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test unsupported options:
            with self.assertRaises(TypeError) as cm:
                client.SSLClient(sslctx, address, foo=17)
            self.assertEqual(str(cm.exception),
                'unsupported SSLClient() **options: foo'
            )
            with self.assertRaises(TypeError) as cm:
                client.SSLClient(sslctx, address, foo=17, bar=18)
            self.assertEqual(str(cm.exception),
                'unsupported SSLClient() **options: bar, foo'
            )

    def test_repr(self):
        class Custom(client.SSLClient):
            pass

        pki = TempPKI()
        sslctx = client.build_client_sslctx(pki.client_sslconfig)
        for address in GOOD_ADDRESSES:
            inst = client.SSLClient(sslctx, address)
            self.assertEqual(repr(inst),
                'SSLClient(<sslctx>, {!r})'.format(address)
            )
            inst = Custom(sslctx, address)
            self.assertEqual(repr(inst),
                'Custom(<sslctx>, {!r})'.format(address)
            )




import sys
from os import path

tree = path.dirname(path.dirname(path.abspath(__file__)))
sys.path.insert(0, tree)

import degu


# Project info
project = 'Degu'
copyright = '2014-2016, Novacut Inc'
version = degu.__version__
release = version


# General config
needs_sphinx = '1.1'
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build']
pygments_style = 'sphinx'


# HTML config
html_theme = 'default'
html_static_path = ['_static']
htmlhelp_basename = 'Degudoc'




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests the `degu.misc` module.
"""

from unittest import TestCase
from os import path

from degu import misc


class TestTempPKI(TestCase):
    def test_init(self):
        # Test when client_pki=True:
        pki = misc.TempPKI()
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'cert')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_id, 'cert')), 0)

        # pki.server_sslconfig property:
        self.assertEqual(pki.server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'ca_file': pki.path(pki.client_ca_id, 'ca'),
        })

        self.assertEqual(pki.anonymous_server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'allow_unauthenticated_clients': True,
        })

        self.assertEqual(pki.client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
            'cert_file': pki.path(pki.client_id, 'cert'),
            'key_file': pki.path(pki.client_id, 'key'),
        })

        self.assertEqual(pki.anonymous_client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
        })

        self.assertTrue(path.isdir(pki.ssldir))
        pki.__del__()
        self.assertFalse(path.exists(pki.ssldir))
        pki.__del__()

        # Test when client_pki=False:
        pki = misc.TempPKI(client_pki=False)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'cert')), 0)

        self.assertEqual(pki.anonymous_server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'allow_unauthenticated_clients': True,
        })

        self.assertEqual(pki.anonymous_client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
        })

        self.assertTrue(path.isdir(pki.ssldir))
        pki.__del__()
        self.assertFalse(path.exists(pki.ssldir))
        pki.__del__()




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
`degu` - an embedded HTTP server and client library.
"""

__version__ = '0.15.0'


# Common IPv6 and IPv6 *address* constants:
IPv6_LOOPBACK = ('::1', 0, 0, 0)
IPv6_ANY = ('::', 0, 0, 0)
IPv4_LOOPBACK = ('127.0.0.1', 0)
IPv4_ANY = ('0.0.0.0', 0)

# Handy for unit testing through *address* permutations:
ADDRESS_CONSTANTS = (
    IPv6_LOOPBACK,
    IPv6_ANY,
    IPv4_LOOPBACK,
    IPv4_ANY,
)


def _run_server(q, address, build_func, *build_args, **options):
    try:
        from .server import Server
        app = build_func(*build_args)
        httpd = Server(address, app, **options)
        q.put(httpd.address)
        httpd.serve_forever()
    except Exception as e:
        q.put(e)
        raise e


def _run_sslserver(q, sslconfig, address, build_func, *build_args, **options):
    try:
        from .server import SSLServer
        app = build_func(*build_args)
        httpd = SSLServer(sslconfig, address, app, **options)
        q.put(httpd.address)
        httpd.serve_forever()
    except Exception as e:
        q.put(e)
        raise e


def _start_server(address, build_func, *build_args, **options):
    import multiprocessing
    q = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=_run_server,
        args=(q, address, build_func) + build_args,
        kwargs=options,
        daemon=True,
    )
    process.start()
    address = q.get()
    if isinstance(address, Exception):
        process.terminate()
        process.join()
        raise address
    return (process, address)


def _start_sslserver(sslconfig, address, build_func, *build_args, **options):
    import multiprocessing
    if not isinstance(sslconfig, dict):
        raise TypeError(
            '{}: need a {!r}; got a {!r}: {!r}'.format(
                'sslconfig', dict, type(sslconfig), sslconfig
            )
        )
    q = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=_run_sslserver,
        args=(q, sslconfig, address, build_func) + build_args,
        kwargs=options,
        daemon=True,
    )
    process.start()
    address = q.get()
    if isinstance(address, Exception):
        process.terminate()
        process.join()
        raise address
    return (process, address)


class _EmbeddedProcess:
    def __del__(self):
        self.terminate()

    def terminate(self):
        if getattr(self, 'process', None) is not None:
            self.process.terminate()
            self.process.join()


class EmbeddedServer(_EmbeddedProcess):
    def __init__(self, address, build_func, *build_args, **options):
        (self.process, self.address) = _start_server(
            address, build_func, *build_args, **options
        )
        self.build_func = build_func
        self.build_args = build_args
        self.options = options


class EmbeddedSSLServer(_EmbeddedProcess):
    def __init__(self, sslconfig, address, build_func, *build_args, **options):
        self.sslconfig = sslconfig
        (self.process, self.address) = _start_sslserver(
            sslconfig, address, build_func, *build_args, **options
        )
        self.build_func = build_func
        self.build_args = build_args
        self.options = options
        self.options = options




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Helpers for non-interactive creation of SSL certs.

This is an extremely trimmed-down version of the Dmedia Identity framework:

    http://bazaar.launchpad.net/~dmedia/dmedia/trunk/view/head:/dmedia/identity.py

Unlike the rest of Degu, this module isn't meant for production use and is
really only aimed at unit testing.

With time, we may move the Dmedia Identity framework into Degu, or at least a
slightly more generalized version of it.  But for now, we're focused on refining
the Degu server and client.

Note that at least for now, we want Degu to only depend on the Python standard
library, so we're using RFC-3548 Base32 encoding for the intrinsic IDs (content
hash of the RSA public key), rather than using Dbase32:

    https://launchpad.net/dbase32

However, be warned that the Dbase32 encoding plays a key role in our long-term
database plans for Novacut and Dmedia, so it might not always be possible to use
RFC-3548 Base32 encoding in that context.
"""

import os
from os import path
from subprocess import check_output
from hashlib import sha512
from base64 import b32encode


DAYS = 365 * 10  # Valid for 10 years


def safe_write(filename, data):
    with open(filename, 'xb', 0) as fp:
        os.chmod(fp.fileno(), 0o400)
        fp.write(data)


def b32enc(data):
    """
    Similar to `dbase32.db32enc()`, but using RFC-3548 Base32 encoding.
    """
    assert 5 <= len(data) <= 60 and len(data) % 5 == 0
    return b32encode(data).decode()


def random_id(numbytes=15):
    """
    Similar to `dbase32.random_id()`, but using RFC-3548 Base32 encoding.
    """
    assert 5 <= numbytes <= 60 and numbytes % 5 == 0
    return b32enc(os.urandom(numbytes))


def hash_pubkey(pubkey_data):
    """
    Hash an RSA public key to compute its intrinsic ID.

    For example:

    >>> hash_pubkey(b'The PEM encoded public key')
    '6RK4BB3YAIHDDJNOFOABZDS4WNEESGEW6LBRPH6AQETUNOMG'

    """
    digest = sha512(pubkey_data).digest()
    return b32enc(digest[:30])


def make_subject(cn):
    """
    Make an openssl certificate subject from the common name *cn*.

    For example:

    >>> make_subject('foo')
    '/CN=foo'

    """
    return '/CN={}'.format(cn)


def create_key(bits):
    """
    Generate new RSA keypair.
    """
    assert bits in (1024, 2048, 3072, 4096)
    return check_output(['openssl', 'genrsa', str(bits)])


def create_ca(key_file, subject):
    """
    Create a self-signed X509 certificate authority.

    *subject* should be an str in the form ``'/CN=foo'``.
    """
    return check_output(['openssl', 'req',
        '-new',
        '-x509',
        '-sha384',
        '-days', str(DAYS),
        '-key', key_file,
        '-subj', subject,
    ])


def create_csr(key_file, subject):
    """
    Create a certificate signing request.

    *subject* should be an str in the form ``'/CN=foo'``.
    """
    return check_output(['openssl', 'req',
        '-new',
        '-sha384',
        '-key', key_file,
        '-subj', subject,
    ])


def issue_cert(csr_file, ca_file, key_file, srl_file):
    """
    Create a signed certificate from a certificate signing request.
    """
    return check_output(['openssl', 'x509',
        '-req',
        '-sha384',
        '-days', str(DAYS),
        '-CAcreateserial',
        '-in', csr_file,
        '-CA', ca_file,
        '-CAkey', key_file,
        '-CAserial', srl_file,
    ])


def get_pubkey(key_data):
    return check_output(['openssl', 'rsa', '-pubout'], input=key_data)


def get_csr_pubkey(csr_data):
    cmd = ['openssl', 'req', '-pubkey', '-noout']
    return check_output(cmd, input=csr_data)


def get_cert_pubkey(cert_data):
    cmd = ['openssl', 'x509', '-pubkey', '-noout']
    return check_output(cmd, input=cert_data)


class PKI:
    __slots__ = ('ssldir',)

    def __init__(self, ssldir):
        self.ssldir = ssldir

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.ssldir)

    def path(self, _id, ext):
        return path.join(self.ssldir, '.'.join([_id, ext]))

    def create_key(self, bits=4096):
        key_data = create_key(bits)
        pubkey_data = get_pubkey(key_data)
        _id = hash_pubkey(pubkey_data)    
        key_file = self.path(_id, 'key')
        safe_write(key_file, key_data)
        return _id

    def create_ca(self, _id):
        key_file = self.path(_id, 'key')
        subject = make_subject(_id)
        ca_data = create_ca(key_file, subject)
        ca_file = self.path(_id, 'ca')
        safe_write(ca_file, ca_data)
        return ca_file

    def create_csr(self, _id):
        key_file = self.path(_id, 'key')
        subject = make_subject(_id)
        csr_data = create_csr(key_file, subject)
        csr_file = self.path(_id, 'csr')
        safe_write(csr_file, csr_data)
        return csr_file

    def issue_cert(self, _id, ca_id):
        csr_file = self.path(_id, 'csr')
        ca_file = self.path(ca_id, 'ca')
        key_file = self.path(ca_id, 'key')
        srl_file = self.path(ca_id, 'srl')
        cert_data = issue_cert(csr_file, ca_file, key_file, srl_file)
        cert_file = self.path(_id, 'cert')
        safe_write(cert_file, cert_data)
        return cert_file

    def get_server_sslconfig(self, server_id, client_ca_id):
        return {
            'cert_file': self.path(server_id, 'cert'),
            'key_file': self.path(server_id, 'key'),
            'ca_file': self.path(client_ca_id, 'ca'),
        }

    def get_anonymous_server_sslconfig(self, server_id):
        """
        Server will accept connections from unauthenticated client.
        """
        return {
            'cert_file': self.path(server_id, 'cert'),
            'key_file': self.path(server_id, 'key'),
            'allow_unauthenticated_clients': True,
        }

    def get_client_sslconfig(self, server_ca_id, client_id):
        return {
            'ca_file': self.path(server_ca_id, 'ca'),
            'check_hostname': False,
            'cert_file': self.path(client_id, 'cert'),
            'key_file': self.path(client_id, 'key'),
        }

    def get_anonymous_client_sslconfig(self, server_ca_id):
        return {
            'ca_file': self.path(server_ca_id, 'ca'),
            'check_hostname': False,
        }




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
RGI validation middleware.

The `Validator` class is a middleware component for verifying that both server
and application comply with the REST Gateway Interface (RGI) specification.

The aim is to be strict and comprehensive, and to deliver exceedingly clear
error messages when non-conforming behavior is detected.

As such, performance is sacrificed for the sake of clarity and maintainability.
The `Validator` middleware is not intended for everyday production use.
Whenever you use it, expect a substantial performance hit.

`degu.rgi` and its tests should be fully self-contained and should not rely on
any other `degu` functionality.  With time, assuming RGI gains wider adoption,
`degu.rgi` should be split out of Degu and into its own source tree, so that it
can better act as an independent RGI validation tool.
"""

# Provide very clear TypeError messages:
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

# Allowed values for request.method:
REQUEST_METHODS = ('GET', 'PUT', 'POST', 'DELETE', 'HEAD')

# 'content-length' and 'transfer-encoding' header keys, defined as constants
# here as typing the literals over and over is rather error prone plus is a bit
# too verbose for comfort at times:
LENGTH = 'content-length'
ENCODING = 'transfer-encoding'


def _getattr(label, obj, name):
    """
    `getattr()` with a clearer error message when attribute is missing.
    """
    if not hasattr(obj, name):
        raise ValueError(
            '{}: {!r} object has no attribute {!r}'.format(
                label, type(obj).__name__, name
            )
        )
    label = '{}.{}'.format(label, name)
    return (label, getattr(obj, name))


def _check_type(label, obj, _type):
    if type(obj) is not type:
        raise TypeError(
            TYPE_ERROR.format(label, _type, type(obj), obj) 
        )


def _ensure_attr_is(label, obj, name, expected):
    """
    Raise a ValueError if *obj* attribute *name* is not *expected*.

    For example, when *obj* has the attribute *name*, but said attribute is not
    *expected*:

    >>> import io
    >>> body = io.BytesIO()
    >>> _ensure_attr_is('request.body', body, 'closed', True)
    Traceback (most recent call last):
      ...
    ValueError: request.body.closed must be True; got False

    Or when *obj* has no attribute *name*:

    >>> _ensure_attr_is('request.body', body, 'chunked', False)
    Traceback (most recent call last):
      ...
    ValueError: request.body: 'BytesIO' object has no attribute 'chunked'

    """
    (label, value) = _getattr(label, obj, name)
    if value is not expected:
        raise ValueError(
            '{} must be {!r}; got {!r}'.format(label, expected, value)
        )


def _ensure_attr_equals(label, obj, name, expected):
    (label, value) = _getattr(label, obj, name)
    if value != expected:
        raise ValueError(
            '{} must equal {!r}; got {!r}'.format(label, expected, value)
        )


def _check_dict(label, obj):
    """
    Ensure that *obj* is a `dict` instance and contains only `str` keys.

    For example, when *obj* isn't a `dict`:

    >>> _check_dict('headers', [('foo', 'bar')])
    Traceback (most recent call last):
      ...
    TypeError: headers: need a <class 'dict'>; got a <class 'list'>: [('foo', 'bar')]

    Or when *obj* contains a non-string key:

    >>> _check_dict('headers', {b'foo': 'bar'})
    Traceback (most recent call last):
      ...
    TypeError: headers: keys must be <class 'str'>; got a <class 'bytes'>: b'foo'

    """
    if not isinstance(obj, dict):
        raise TypeError(TYPE_ERROR.format(label, dict, type(obj), obj))
    for key in obj:
        if not isinstance(key, str):
            raise TypeError('{}: keys must be {!r}; got a {!r}: {!r}'.format(
                    label, str, type(key), key
                )
            )


def _check_headers(label, headers):
    """
    Validate *headers* dictionary.

    `_validate_request()` uses this to validate the incoming request headers.

    `_validate_response()` uses this to validate the outgoing response headers.
    """
    _check_dict(label, headers)
    for (key, value) in headers.items():
        if key != key.casefold():
            raise ValueError(
                '{}: non-casefolded header name: {!r}'.format(label, key)
            )
        if not isinstance(value, str) and key != LENGTH:
            raise TypeError(
                '{}[{!r}]: need a {!r}; got a {!r}: {!r}'.format(
                    label, key, str, type(value), value
                )
            )
    if LENGTH in headers:
        if ENCODING in headers:
            raise ValueError(
                '{}: {} and {} in headers'.format(label, LENGTH, ENCODING)
            )
        (l, v) = _get_path(label, headers, LENGTH)
        if not isinstance(v, int):
            raise TypeError(TYPE_ERROR.format(l, int, type(v), v))
        if v < 0:
            raise ValueError('{}: must be >=0; got {!r}'.format(l, v))
    if ENCODING in headers:
        (l, v) = _get_path(label, headers, ENCODING)
        if v != 'chunked':
            raise ValueError("{}: must be 'chunked'; got {!r}".format(l, v))


def _check_address(label, address):
    """
    Validate a socket address.

    `_validate_session()` uses this to validate ``session.address``.

    For example:

    >>> _check_address('session.address', ('::1', 1234, 0, 0))
    >>> _check_address('session.address', ('::1', 1234, 0))
    Traceback (most recent call last):
      ...
    ValueError: session.address: tuple must have 2 or 4 items; got ('::1', 1234, 0)

    """
    if isinstance(address, tuple):
        if len(address) not in (2, 4):
            raise ValueError(
                '{}: tuple must have 2 or 4 items; got {!r}'.format(label, address)
            )
    elif not isinstance(address, (str, bytes)):
        raise TypeError(
            TYPE_ERROR.format(label, (tuple, str, bytes), type(address), address)
        )


def _get_path(label, value, *path):
    """
    Return a ``(label, value)`` tuple.

    For example, with an empty path:

    >>> session = {'client': ('127.0.0.1', 52521)}
    >>> _get_path('session', session)
    ('session', {'client': ('127.0.0.1', 52521)})

    Or with a single value path:

    >>> _get_path('session', session, 'client')
    ("session['client']", ('127.0.0.1', 52521))

    Or with a path that is 2 deep:

    >>> _get_path('session', session, 'client', 0)
    ("session['client'][0]", '127.0.0.1')
    >>> _get_path('session', session, 'client', 1)
    ("session['client'][1]", 52521)

    Or when first path item is missing:

    >>> _get_path('session', session, 'foo')
    Traceback (most recent call last):
      ...
    ValueError: session['foo'] does not exist

    Or when the 2nd path item is missing:

    >>> _get_path('session', session, 'client', 2)
    Traceback (most recent call last):
      ...
    ValueError: session['client'][2] does not exist

    Note that this function carries a substantial performance overhead.  But the
    point is to be clear, correct, and maintainable, so that's okay :D
    """
    for key in path:
        assert isinstance(key, (str, int))
        label = '{}[{!r}]'.format(label, key)
        try:
            value = value[key]
        except (KeyError, IndexError):
            raise ValueError(
                '{} does not exist'.format(label)
            )
    return (label, value)


def _validate_session(session):
    """
    Validate the *session* argument.
    """

    # address:
    (label, value) = _getattr('session', session, 'address')
    _check_address(label, value)

    # requests:
    (label, value) = _getattr('session', session, 'requests')
    if not isinstance(value, int):
        raise TypeError(
            TYPE_ERROR.format(label, int, type(value), value) 
        )
    if value < 0:
        raise ValueError('{} must be >= 0; got {!r}'.format(label, value))

    # store:
    (label, value) = _getattr('session', session, 'store')
    _check_dict(label, value)

    # FIXME: Add tests for session.credentials, session.max_requests


def _reconstruct_uri(request):
    uri = '/' + '/'.join(request.mount + request.path)
    if request.query is None:
        return uri
    return '?'.join([uri, request.query])


def _check_uri_invariant(request):
    uri = _reconstruct_uri(request)
    if uri != request.uri:
        raise ValueError(
            "reconstruct_uri(request) != request.uri: {!r} != {!r}".format(
                uri, request.uri
            )
        )


def _validate_request(bodies, request):
    """
    Validate the *request* argument.
    """

    # request.method:
    (label, value) = _getattr('request', request, 'method')
    if value not in REQUEST_METHODS:
        raise ValueError(
            "{}: value {!r} not in {!r}".format(label, value, REQUEST_METHODS)
        )
    if not (getattr(request, 'body', None) is None or value in {'PUT', 'POST'}):
        raise ValueError(
            "{} cannot be {!r} when request.body is not None".format(label, value)
        )

    # request.uri:
    (label, value) = _getattr('request', request, 'uri')
    if not isinstance(value, str):
        raise TypeError(
            TYPE_ERROR.format(label, str, type(value), value) 
        )

    # request.mount:
    (label, value) = _getattr('request', request, 'mount')
    if not isinstance(value, list):
        raise TypeError(
            TYPE_ERROR.format(label, list, type(value), value) 
        )
    for i in range(len(value)):
        (l, v) = _get_path(label, value, i)
        if not isinstance(v, str):
            raise TypeError(TYPE_ERROR.format(l, str, type(v), v))

    # request.path:
    (label, value) = _getattr('request', request, 'path')
    if not isinstance(value, list):
        raise TypeError(
            TYPE_ERROR.format(label, list, type(value), value) 
        )
    for i in range(len(value)):
        (l, v) = _get_path(label, value, i)
        if not isinstance(v, str):
            raise TypeError(TYPE_ERROR.format(l, str, type(v), v))

    # request.query:
    (label, value) = _getattr('request', request, 'query')
    if not (value is None or isinstance(value, str)):
        raise TypeError(
            TYPE_ERROR.format(label, str, type(value), value) 
        )

    # request.headers:
    (label, value) = _getattr('request', request, 'headers')
    _check_headers(label, value)

    # request.body:
    (label, value) = _getattr('request', request, 'body')
    if value is None:
        # When there is no request body, there should be neither 'content-length'
        # nor 'tranfer-encoding' headers:
        for key in (LENGTH, ENCODING):
            if key in request.headers:
                raise ValueError(
                    "{} is None but {!r} header is included".format(label, key)
                )
        return  # Skip remaining checks when request body is None

    # request.body is not None:
    assert value is not None
    if isinstance(value, bodies.Body):
        _ensure_attr_is(label, value, 'chunked', False)
        if ENCODING in request.headers:
            raise ValueError(
                "{}: bodies.Body with {!r} header".format(label, ENCODING)
            )
        (l1, v1) = _getattr(label, value, 'content_length')
        if LENGTH not in request.headers:
            raise ValueError(
                "{}: bodies.Body, but missing {!r} header".format(label, LENGTH)
            )
        (l2, v2) = _get_path('request.headers', request.headers, LENGTH)
        if v1 != v2:
            raise ValueError(
                '{} != {}: {!r} != {!r}'.format(l1, l2, v1, v2)
            )
    elif isinstance(value, bodies.ChunkedBody):
        _ensure_attr_is(label, value, 'chunked', True)
        if LENGTH in request.headers:
            raise ValueError(
                "{}: bodies.ChunkedBody with {!r} header".format(label, LENGTH)
            )
        if ENCODING not in request.headers:
            raise ValueError(
                "{}: bodies.ChunkedBody, but missing {!r} header".format(label, ENCODING)
            )
        assert request.headers[ENCODING] == 'chunked'
    else:
        body_types = (bodies.Body, bodies.ChunkedBody)
        raise TypeError(
            TYPE_ERROR.format(label, body_types, type(value), value) 
        )

    # Need request.body.state == 0 prior to calling the application:
    _ensure_attr_equals(label, value, 'state', 0)


def _validate_response(bodies, request, response):
    """
    Validate the *response* tuple returned by the application.
    """
    if not isinstance(response, tuple):
        raise TypeError(
            TYPE_ERROR.format('response', tuple, type(response), response)
        )
    if len(response) != 4:
        raise ValueError(
            'len(response) must be 4, got {}'.format(len(response))
        )

    # response[0] (status):
    (label, value) = _get_path('response', response, 0)
    if not isinstance(value, int):
        raise TypeError(
            TYPE_ERROR.format(label, int, type(value), value)
        )
    if not (100 <= value <= 599):
        raise ValueError(
            '{}: need 100 <= status <= 599; got {}'.format(label, value)
        )

    # response[1] (reason):
    (label, value) = _get_path('response', response, 1)
    if not isinstance(value, str):
        raise TypeError(
            TYPE_ERROR.format(label, str, type(value), value)
        )
    if value == '':
        raise ValueError(
            "{}: reason cannot be an empty ''".format(label)
        )

    # response[2] (headers):
    (label, value) = _get_path('response', response, 2)
    _check_headers(label, value)
    if request.method == 'HEAD':
        # response to 'HEAD' request must include either a 'content-length' or a
        # 'transfer-encoding' header (but not both):
        if not {LENGTH, ENCODING}.intersection(value):
            raise ValueError(
                "{}: response to HEAD request must include {!r} or {!r} header".format(
                    label, LENGTH, ENCODING
                )
            )

    # response[3] (body):
    (label, value) = _get_path('response', response, 3)
    if request.method == 'HEAD':
        # response body must be None when request method is 'HEAD':
        if value is not None:
            raise TypeError(
                "{}: must be None when request.method is 'HEAD'; got a {!r}".format(
                    label, type(value)
                )
            )
        return  # Skip remaining tests
    if value is None:
        # When response body is None and request method is not 'HEAD', should
        # include neither 'content-length' nor 'transfer-encoding' headers:
        for key in (LENGTH, ENCODING):
            if key in response[2]:
                raise ValueError(
                    '{}: response body is None, but {!r} header is included'.format(label, key)
                )
        return  # Skip remaning tests
    length_types = (bytes, bytearray, bodies.Body, bodies.BodyIter)
    chunked_types = (bodies.ChunkedBody, bodies.ChunkedBodyIter)
    if isinstance(value, length_types):
        # Cannot include a 'transfer-encoding' header with a length-encoded
        # response body:
        if ENCODING in response[2]:
            raise ValueError(
                '{}: response body is {!r}, but {!r} header is included'.format(
                    label, type(value), ENCODING
                )
            )
        if isinstance(value, (bytes, bytearray)):
            length = len(value)
            _repr = 'len(body)'
        else:
            (l, length) = _getattr(label, value, 'content_length')
            _repr = 'body.content_length'
            _ensure_attr_is(label, value, 'chunked', False)
            _ensure_attr_equals(label, value, 'state', 0)
        if LENGTH in response[2] and response[2][LENGTH] != length:
            raise ValueError(
                '{}: {} is {}, but {!r} is {}'.format(
                    label, _repr, length, LENGTH, response[2][LENGTH]
                )
            )
    elif isinstance(value, chunked_types):
        # Cannot include a 'content-length' header with a chunk-encoded response
        # body:
        if LENGTH in response[2]:
            raise ValueError(
                '{}: response body is {!r}, but {!r} header is included'.format(
                    label, type(value), LENGTH
                )
            )
        _ensure_attr_is(label, value, 'chunked', True)
        _ensure_attr_equals(label, value, 'state', 0)
        name = 'content_length'
        if hasattr(value, name):
            raise ValueError(
                '{}: {!r} must not have a {!r} attribute'.format(
                    label, type(value), name
                )
            )
    else:
        raise TypeError(
            '{}: bad response body type: {!r}'.format(label, type(value))
        )


class Validator:
    __slots__ = ('app', '_on_connect')

    def __init__(self, app):
        if not callable(app):
            raise TypeError('app: not callable: {!r}'.format(app))
        on_connect = getattr(app, 'on_connect', None)
        if not (on_connect is None or callable(on_connect)):
            raise TypeError(
                'app.on_connect: not callable: {!r}'.format(on_connect)
            )
        self.app = app
        self._on_connect = on_connect

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.app)

    def __call__(self, session, request, bodies):
        _validate_session(session)
        _validate_request(bodies, request)
        _check_uri_invariant(request)
        response = self.app(session, request, bodies)
        _check_uri_invariant(request)
        if request.body is not None and request.body.state != 2:
            # request body was not fully consumed:
            raise ValueError(
                '{} must be 2 after app() was called; got {!r}'.format(
                    "request.body.state", request.body.state
                )
            )
        _validate_response(bodies, request, response)
        return response

    def on_connect(self, session, sock):
        _validate_session(session)
        if session.requests != 0:
            raise ValueError(
                '{} must be 0 when app.on_connect() is called; got {!r}'.format(
                    'session.requests', session.requests
                )
            )
        if self._on_connect is None:
            return True
        allow = self._on_connect(session, sock)
        if not isinstance(allow, bool):
            raise TypeError(
                'app.on_connect() must return a {!r}; got a {!r}: {!r}'.format(
                    bool, type(allow), allow
                )
            )
        return allow




#!/usr/bin/env python3

# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Install `degu`.
"""

import sys
if sys.version_info < (3, 4):
    sys.exit('ERROR: degu requires Python 3.4 or newer')

import os
from os import path
import platform
import subprocess
from distutils.core import setup, Extension
from distutils.cmd import Command

import degu
from degu.tests.run import run_tests


TREE = path.dirname(path.abspath(__file__))
with open(path.join(TREE, 'README'), 'r', encoding='utf-8') as fp:
    LONG_DESCRIPTION = fp.read()


def run_under_same_interpreter(opname, script, args):
    print('\n** running: {}...'.format(script), file=sys.stderr)
    if not os.access(script, os.R_OK | os.X_OK):
        print('ERROR: cannot read and execute: {!r}'.format(script),
            file=sys.stderr
        )
        print('Consider running `setup.py test --skip-{}`'.format(opname),
            file=sys.stderr
        )
        sys.exit(3)
    cmd = [sys.executable, script] + args
    print('check_call:', cmd, file=sys.stderr)
    subprocess.check_call(cmd)
    print('** PASSED: {}\n'.format(script), file=sys.stderr)


def run_sphinx_doctest():
    script = '/usr/share/sphinx/scripts/python3/sphinx-build'
    doc = path.join(TREE, 'doc')
    doctest = path.join(TREE, 'doc', '_build', 'doctest')
    args = ['-EW', '-b', 'doctest', doc, doctest]
    run_under_same_interpreter('sphinx', script, args)


def run_pyflakes3():
    script = '/usr/bin/pyflakes3'
    names = [
        'degu',
        'setup.py',
        'benchmark.py',
        'benchmark-parsing.py',
        'benchmark-ssl.py',
    ]
    args = [path.join(TREE, name) for name in names]
    run_under_same_interpreter('flakes', script, args)


class Test(Command):
    description = 'run unit tests and doctests'

    user_options = [
        ('skip-sphinx', None, 'do not run Sphinx doctests'),
        ('skip-flakes', None, 'do not run pyflakes static checks'),
        ('skip-slow', None, 'skip the rather slow socket timeout tests'),
    ]

    def initialize_options(self):
        self.skip_sphinx = 0
        self.skip_flakes = 0
        self.skip_slow = 0

    def finalize_options(self):
        pass

    def run(self):
        if self.skip_slow:
            os.environ['DEGU_TEST_SKIP_SLOW'] = 'true'
        if not run_tests():
            sys.exit(2)
        if not self.skip_sphinx:
            run_sphinx_doctest()
        if not self.skip_flakes:
            run_pyflakes3()


ext_kw = {
    'sources': ['degu/_base.c'],
    'extra_compile_args': [
        '-Werror',  # Make warnings into errors
        '-Wall',
        '-Wsign-compare',
        '-Wsign-conversion',
        '-Wmissing-field-initializers',
        '-Wfatal-errors',
        '-std=gnu11',
        '-pedantic-errors',
        '-Wpedantic',
    ],
}
if os.environ.get('DEGU_INSTRUMENT_BUILD') == 'true':
    ext_kw['extra_compile_args'].append('-fsanitize=address')
    ext_kw['libraries'] = ['asan']
    if platform.dist()[1] >= '15.04':
        ext_kw['extra_compile_args'].extend([
            '-fsanitize=undefined',
            '-fsanitize=shift',
            '-fsanitize=integer-divide-by-zero',
            '-fsanitize=unreachable',
            '-fsanitize=vla-bound',
            '-fsanitize=null',
            '-fsanitize=return',
            '-fsanitize=signed-integer-overflow',
        ])
        ext_kw['libraries'].append('ubsan')


setup(
    name='degu',
    description='Embedded HTTP server and client library',
    long_description=LONG_DESCRIPTION,
    url='https://launchpad.net/degu',
    version=degu.__version__,
    author='Jason Gerard DeRose',
    author_email='jderose@novacut.com',
    license='LGPLv3+',
    packages=[
        'degu',
        'degu.tests',
    ],
    ext_modules=[
        Extension('degu._base', **ext_kw),
    ],
    cmdclass={
        'test': Test,
    },
    platforms=['POSIX'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit test helpers.
"""

import io
import sys
import os
from os import path
import tempfile
import shutil
from random import SystemRandom
from unittest import TestCase
import string

from degu import tables
from degu.sslhelpers import random_id


MAX_IO_SIZE = 16777216  # 16 MiB
random = SystemRandom()


def random_data():
    """
    Return random bytes between 1 and 34969 (inclusive) bytes long.

    In unit tests, this is used to simulate a random request or response body,
    or a random chunk in a chuck-encoded request or response body.
    """
    size = random.randint(1, 34969)
    return os.urandom(size)


def random_chunks():
    """
    Return between 0 and 10 random chunks (inclusive).

    There will always be 1 additional, final chunk, an empty ``b''``, as per the
    HTTP/1.1 specification.
    """
    count = random.randint(0, 10)
    chunks = [random_data() for i in range(count)]
    chunks.append(b'')
    return chunks


def random_identifier():
    return ''.join(random.choice(string.ascii_lowercase) for i in range(17))


def random_chunk_ext():
    if random.randrange(3) == 0:
        return None
    return (random_identifier(), random_identifier())

def random_chunk(size=None):
    if size is None:
        size = random.randint(1, 34969)
    assert type(size) is int and 0 <= size <= MAX_IO_SIZE
    ext = random_chunk_ext()
    data = os.urandom(size)
    return (ext, data)

def random_chunks2(count=None):
    """
    Return between 0 and 10 random chunks (inclusive).

    There will always be 1 additional, final chunk, an empty ``b''``, as per the
    HTTP/1.1 specification.
    """
    if count is None:
        count = random.randrange(10) + 1
    assert type(count) is int and count > 0
    chunks = [
        random_chunk() for i in range(count - 1)
    ]
    chunks.append(random_chunk(0))
    assert len(chunks) == count
    return tuple(chunks)


def iter_good(good, allowed):
    for i in range(len(good)):
        for g in allowed:
            also_good = bytearray(good)
            also_good[i] = g
            yield bytes(also_good)


def iter_bad(good, allowed):
    assert isinstance(good, bytes)
    assert isinstance(allowed, bytes)
    not_allowed = tables.invert(allowed)
    for i in range(len(good)):
        for b in not_allowed:
            bad = bytearray(good)
            bad[i] = b
            yield bytes(bad)


class TempDir:
    def __init__(self, prefix='unittest.'):
        self.dir = tempfile.mkdtemp(prefix=prefix)

    def __del__(self):
        shutil.rmtree(self.dir)

    def join(self, *parts):
        return path.join(self.dir, *parts)

    def mkdir(self, *parts):
        dirname = self.join(*parts)
        os.mkdir(dirname)
        return dirname

    def makedirs(self, *parts):
        dirname = self.join(*parts)
        os.makedirs(dirname)
        return dirname

    def touch(self, *parts):
        filename = self.join(*parts)
        open(filename, 'xb').close()
        return filename

    def create(self, *parts):
        filename = self.join(*parts)
        return (filename, open(filename, 'xb'))

    def write(self, data, *parts):
        (filename, fp) = self.create(*parts)
        fp.write(data)
        fp.close()
        return filename

    def prepare(self, content):
        filename = self.write(content, random_id())
        return open(filename, 'rb')


class DummySocket:
    __slots__ = ('_calls',)

    def __init__(self):
        self._calls = []

    def shutdown(self, how):
        self._calls.append(('shutdown', how))

    def recv_into(self, buf):
        self._calls.append(('recv_into', buf))

    def send(self, buf):
        self._calls.append(('send', buf))


class DummyFile:
    def __init__(self):
        self._calls = []

    def close(self):
        self._calls.append('close')


class MockSocket:
    __slots__ = ('_rfile', '_wfile', '_rcvbuf', '_recv_into_calls', '_calls')

    def __init__(self, data, rcvbuf=None):
        assert rcvbuf is None or (isinstance(rcvbuf, int) and rcvbuf > 0)
        self._rfile = io.BytesIO(data)
        self._wfile = io.BytesIO()
        self._rcvbuf = rcvbuf
        self._recv_into_calls = 0
        self._calls = []

    def recv_into(self, buf):
        assert isinstance(buf, memoryview)
        self._calls.append(('recv_into', len(buf)))
        if self._rcvbuf is not None and len(buf) > self._rcvbuf:
            buf = buf[0:self._rcvbuf]
        self._recv_into_calls += 1
        return self._rfile.readinto(buf)

    def send(self, data):
        return self._wfile.write(data)


class MockSocket2:
    __slots__ = ('__rfile', '__rcvbuf')

    def __init__(self, rfile, rcvbuf=None):
        assert rcvbuf is None or (type(rcvbuf) is int and rcvbuf > 0)
        self.__rfile = rfile
        self.__rcvbuf = rcvbuf

    def recv_into(self, buf):
        assert isinstance(buf, memoryview)
        if self.__rcvbuf is not None:
            buf = buf[0:self.__rcvbuf]
        return self.__rfile.readinto(buf)


class FuzzTestCase(TestCase):
    """
    Base class for fuzz-testing read functions.
    """

    def fuzz(self, func, *args):
        """
        Perform random fuzz test on *func*.

        Expected result: given an rfile containing 8192 random bytes, func()
        should raise a ValueError every time, should read at least 1 byte, and
        should never read more than 4096 bytes.
        """
        for i in range(1000):
            data = os.urandom(4096 * 2)
            rfile = io.BytesIO(data)
            self.assertEqual(sys.getrefcount(rfile), 2)
            with self.assertRaises(ValueError):
                func(rfile, *args)
            self.assertGreaterEqual(rfile.tell(), 1)
            self.assertLessEqual(rfile.tell(), 4096)
            # Make sure refcount is still correct (especially important for
            # testing C extensions):
            self.assertEqual(sys.getrefcount(rfile), 2)


class MockBodies:
    def __init__(self, **kw):
        for (key, value) in kw.items():
            assert key in ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')
            setattr(self, key, value)


def iter_bodies_with_missing_object():
    names = ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')

    def dummy_body():
        pass

    for name in names:
        kw = dict((key, dummy_body) for key in names)
        del kw[name]
        yield (MockBodies(**kw), name)


def iter_bodies_with_non_callable_object():
    names = ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')

    def dummy_body():
        pass

    for name in names:
        kw = dict((key, dummy_body) for key in names)
        attr = random_identifier()
        kw[name] = attr
        yield (MockBodies(**kw), name, attr)


def build_uri(path, query):
    uri = '/' + '/'.join(path)
    if query is None:
        return uri
    return '?'.join([uri, query])


def iter_random_path():
    yield []
    for length in range(1, 5):
        p = [random_id() for i in range(length)]
        yield p
        yield p + [''] 


def iter_random_uri():
    queries = (
        None,
        random_id(),
        '{}={}'.format(random_id(), random_id()),
        '{}={}&{}={}'.format(random_id(), random_id(), random_id(), random_id())
    )
    for p in iter_random_path():
        for q in queries:
            yield (build_uri(p, q), p, q)





# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Generate tables for validating the HTTP preamble and the HTTP chunk line.

Print the C tables like this::

    $ python3 -m degu.tables

Or print the Python tables like this::

    $ python3 -m degu.tables --python

Finally, replace the existing tables in the "degu/_base.h" and "degu/_basepy.py"
files with the tables as currently defined in this module like this::

    $ python3 -m degu.tables --update

Obviously you should only use --update within the source tree.

If the existing tables already matched the current table definitions, the
content of these files should not change (if it does change, that's a bug).
However, the mtime of these files will change regardless, as a new file is still
written even when the resulting content will be the same. 
"""

from collections import namedtuple
import os
from os import path
import argparse


# Provide very clear TypeError messages:
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'


Table = namedtuple('Table', 'name ignore items')
BitFlag = namedtuple('BitFlag', 'bit name allowed')
BitMask = namedtuple('BitMask', 'mask name flags allowed')
Info = namedtuple('Info', 'flags masks table')
Markers = namedtuple('Markers', 'begin end')
Numbers = namedtuple('Numbers', 'decimal hexadecimal')

COOKIE = b' -/0123456789:;=ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz'

# For case-folding and validating header names:
NAMES_DEF = b'-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'

# For parsing decimal and hexadecimal values:
DECIMAL_DEF = b'0123456789'
HEXADECIMAL_DEF = DECIMAL_DEF + b'ABCDEFabcdef'
NUM_DEF = Numbers(DECIMAL_DEF, HEXADECIMAL_DEF)

# Generic bit-flag based validation table with 7 sets, plus 1 error set:
BIT_FLAGS_DEF = (
    ('LOWER', b'-0123456789abcdefghijklmnopqrstuvwxyz'),
    ('UPPER', b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
    ('URI',   b'/?'),
    ('PATH',  b'+.:_~'),
    ('QUERY', b'%&='),
    ('SPACE', b' '),
    ('VALUE', b'"\'()*,;<>[]'),   
)

# The above can be combined to produce a number of validation masks:
BIT_MASKS_DEF = (
    ('KEY',    ('LOWER',)),
    ('VAL',    ('LOWER', 'UPPER', 'PATH', 'QUERY', 'URI', 'SPACE', 'VALUE')),
    ('URI',    ('LOWER', 'UPPER', 'PATH', 'QUERY', 'URI')),
    ('PATH',   ('LOWER', 'UPPER', 'PATH')),
    ('QUERY',  ('LOWER', 'UPPER', 'PATH', 'QUERY')),
    ('REASON', ('LOWER', 'UPPER', 'SPACE')),
    ('EXTKEY', ('LOWER', 'UPPER')),
    ('EXTVAL', ('LOWER', 'UPPER', 'PATH', 'VALUE')),
)


def _iter_never_allowed():
    for i in range(256):
        if (0 <= i <= 31) or (127 <= i <= 255):
            yield i
        else:
            c = chr(i)
            assert c == bytes([i]).decode('ascii')
            if c.isprintable() is False:
                yield i

NEVER_ALLOWED = bytes(_iter_never_allowed())
NEVER_ALLOWED_SET = frozenset(NEVER_ALLOWED)


def normalize(source):
    return bytes(sorted(set(source)))


def check_allowed(allowed):
    if not isinstance(allowed, bytes):
        raise TypeError(
            TYPE_ERROR.format('allowed', bytes, type(allowed), allowed)
        )
    expected = normalize(allowed)
    if allowed != expected:
        raise ValueError('{!r} != {!r}'.format(allowed, expected))
    if NEVER_ALLOWED_SET.intersection(allowed):
        raise ValueError(
            'contains bytes that are never allowed: {!r}'.format(allowed)
        )
    return allowed


def invert(allowed):
    return normalize(set(range(256)) - set(check_allowed(allowed)))


def check_disjoint(accum, allowed):
    check_allowed(allowed)
    common = set(accum).intersection(allowed)
    if common:
        common = normalize(common)
        accum = normalize(accum)
        raise ValueError(
            '{!r} common between {!r} and {!r}'.format(
                normalize(common), normalize(accum), allowed
            )
        )


def build_names_table(allowed):
    check_allowed(allowed)
    items = []
    for i in range(256):
        if i in allowed:
            r = ord(chr(i).lower())
            pair = (i, r)
        else:
            pair = (i, 255)
        items.append(pair)
    return Table('_NAME', 255, tuple(items))


def build_num_table(num):
    assert type(num) is Numbers
    allowed = num.hexadecimal
    check_allowed(allowed)
    items = []
    for i in range(256):
        if i in allowed:
            r = allowed.index(bytes([i]).upper())
            assert 0 <= r < 16
            if r >= 10:
                r |= 16
            else:
                assert i in num.decimal
            pair = (i, r)
        else:
            pair = (i, 255)
        items.append(pair)
    return Table('_NUMBER', 255, tuple(items))


def build_flags(flags_def):
    assert 0 < len(flags_def) < 8
    accum = []
    for (i, (name, allowed)) in enumerate(flags_def):
        bit = 2 ** i
        assert bit in (1, 2, 4, 8, 16, 32, 64)
        accum.append(BitFlag(bit, name, allowed))
    return tuple(accum)


def build_masks(flags, masks_def):
    accum = []
    _map = dict((f.name, f) for f in flags)
    for (name, parts) in masks_def:
        bits = 0
        union = set()
        for p in parts:
            f = _map[p]
            bits |= f.bit
            check_disjoint(union, f.allowed)
            union.update(f.allowed)
        mask = 255 ^ bits
        allowed = normalize(union)
        assert len(allowed) == len(union)
        accum.append(BitMask(mask, name, parts, allowed))
    return tuple(accum)


def build_flags_table(flags):
    assert 1 < len(flags) < 8
    table = {}
    for f in flags:
        check_allowed(f.allowed)
        assert set(table).isdisjoint(f.allowed)
        table.update((key, f.bit) for key in f.allowed)
    for key in range(256):
        table.setdefault(key, 128)
    assert len(table) == 256
    items = tuple(sorted(table.items()))
    return Table('_FLAG', 128, items)


def build_info(flags_def, masks_def):
    flags = build_flags(flags_def)
    masks = build_masks(flags, masks_def)
    table = build_flags_table(flags)
    return Info(flags, masks, table)


def format_table_row(row):
    return ','.join('{:>3}'.format(r) for (i, r) in row)


def format_table_row_help(row, ignore):
    if set(r for (i, r) in row) == {ignore}:
        return None
    help = []
    for (i, r) in row:
        if r == ignore:
            help.append(' ' * 4)  # 4 spaces
        else:
            help.append('{!r:<4}'.format(chr(i)))
    return ' '.join(help)


def iter_c_table_rows(table):
    assert table.ignore in (255, 128)
    row = []
    for item in table.items:
        row.append(item)
        if len(row) == 8:
            line = '    {},'.format(format_table_row(row))
            help = format_table_row_help(row, table.ignore)
            if help:
                yield '{} //  {}'.format(line, help.rstrip())
            else:
                yield line
            row = []
    assert not row


def iter_c_table(table):
    assert len(table.items) == 256
    yield 'static const uint8_t {}[{:d}] = {{'.format(
        table.name, len(table.items)
    )
    yield from iter_c_table_rows(table)
    yield '};'


def iter_c_bit_flags_comment(flags):
    width = max(len(f.name) for f in flags)
    yield '/*'
    for f in flags:
        name = f.name.ljust(width)
        yield ' * {} {:>2} {:08b}  {!r}'.format(name, f.bit, f.bit, f.allowed)
    yield ' */'


def mask_name(name):
    assert 'mask' not in name.lower()
    assert name.upper() == name
    return name + '_MASK'


def iter_c_bit_masks(masks):
    width = max(len(mask_name(m.name)) for m in masks)
    for m in masks:
        name = mask_name(m.name).ljust(width)
        assert m.flags is not None
        parts = '~({})'.format('|'.join(m.flags))
        yield '#define {} {:>3}  // {:08b} {}'.format(
            name, m.mask, m.mask, parts
        )


def iter_c_info(info):
    yield from iter_c_bit_flags_comment(info.flags)
    yield from iter_c_bit_masks(info.masks)
    yield from iter_c_table(info.table)


def py_flag_name(name):
    return '_{}'.format(name)


def iter_py_info(info):
    width = max(len(py_flag_name(f.name)) for f in info.flags)
    for f in info.flags:
        name = py_flag_name(f.name).ljust(width)
        yield '{} = {!r}'.format(name, f.allowed)
    yield ''
    width = max(len(m.name) for m in info.masks)
    for m in info.masks:
        name = m.name.ljust(width)
        src = ' + '.join(py_flag_name(n) for n in m.flags)
        yield '{} = frozenset({})'.format(name, src)


def build_marker_comments(end, fill):
    labels = tuple(
        '{} GENERATED TABLES'.format(way) for way in ('BEGIN', 'END')
    )
    width = max(len(l) for l in labels)
    markers = []
    for label in labels:
        line = ''.join([
            end,
            (fill * 15),
            (' ' * 4),
            label.ljust(width),
            (' ' * 4),
        ])
        needfill = 80 - len(line) - 1
        line += (fill * needfill) + end
        markers.append(line)
    return Markers(*markers)


def replace(inlines, markers, newlines):
    # States:
    #   0: start marker not yet found
    #   1: start marker found, end marker not yet found
    #   2: end marker found (no markers should be found again)
    state = 0
    outlines = []
    for line in inlines:
        assert state in (0, 1, 2)
        if line == markers.begin:
            assert state == 0
            state += 1
        if state in (0, 2):
            outlines.append(line)
        if line == markers.end:
            assert state == 1
            state += 1
            for new in newlines:
                outlines.append(new)
    assert state == 2
    return outlines


def update(pkgdir, name, markers, newlines):
    orig = path.join(pkgdir, name)
    tmp = orig + '.updated-tables.new'
    bak = orig + '.updated-tables.old'
    text = open(orig, 'r').read()
    inlines = text.splitlines()
    outlines = replace(inlines, markers, newlines)
    with open(tmp, 'x') as fp:
        fp.write('\n'.join(outlines) + '\n')
    os.rename(orig, bak)
    os.rename(tmp, orig)
    print('Updated {!r}'.format(orig))


class Generated:
    def __init__(self, names_def, flags_def, masks_def, num_def):
        self.names_def = names_def
        self.names_table = build_names_table(names_def)
        self.info = build_info(flags_def, masks_def)
        self.markers_c = build_marker_comments('/', '*')
        self.markers_py = build_marker_comments('#', '#')
        self.num_def = num_def
        self.num_table = build_num_table(num_def)

    def iter_lines_c(self):
        yield self.markers_c.begin
        yield from iter_c_table(self.names_table)
        yield ''
        yield from iter_c_table(self.num_table)
        yield ''
        yield from iter_c_info(self.info)
        yield self.markers_c.end

    def iter_lines_py(self):
        yield self.markers_py.begin
        yield 'NAME = frozenset('
        yield '    {!r}'.format(self.names_def)
        yield ')'
        yield ''
        yield 'DECIMAL = frozenset({!r})'.format(self.num_def.decimal)
        yield 'HEXADECIMAL = frozenset({!r})'.format(self.num_def.hexadecimal)
        yield ''
        yield from iter_py_info(self.info)
        yield self.markers_py.end

    def update(self, pkgdir):
        update(pkgdir, '_base.h', self.markers_c, self.iter_lines_c())
        update(pkgdir, '_basepy.py', self.markers_py, self.iter_lines_py())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--python', action='store_true', default=False,
        help='print Python tables (instead of C)'
    )
    parser.add_argument('--update', action='store_true', default=False,
        help="update tables in 'degu/_base.c', 'degu/_basepy.py'"
    )
    args = parser.parse_args()

    gen = Generated(NAMES_DEF, BIT_FLAGS_DEF, BIT_MASKS_DEF, NUM_DEF)
    if args.update:
        pkgdir = path.dirname(path.abspath(__file__))
        gen.update(pkgdir)
    else:
        if args.python:
            lines = gen.iter_lines_py()
        else:
            lines = gen.iter_lines_c()
        for line in lines:
            print(line)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
HTTP client.
"""

import os
import socket

from .base import Connection, Response, _TYPE_ERROR


__all__ = ('Client', 'SSLClient', 'Connection', 'Response')


def build_client_sslctx(sslconfig):
    """
    Build an `ssl.SSLContext` appropriately configured for client-side use.

    For example:

    >>> sslconfig = {
    ...     'check_hostname': False,
    ...     'ca_file': '/my/server.ca',
    ...     'cert_file': '/my/client.cert',
    ...     'key_file': '/my/client.key',
    ... }
    >>> sslctx = build_client_sslctx(sslconfig)  #doctest: +SKIP

    """
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if not isinstance(sslconfig, dict):
        raise TypeError(
            _TYPE_ERROR.format('sslconfig', dict, type(sslconfig), sslconfig)
        )

    # In typical Degu P2P usage, hostname checking is meaningless because we
    # wont be trusting centralized certificate authorities, and will typically
    # only connect to servers via their IP address; however, it's still prudent
    # to make *check_hostname* default to True:
    check_hostname = sslconfig.get('check_hostname', True)
    if not isinstance(check_hostname, bool):
        raise TypeError(_TYPE_ERROR.format(
            "sslconfig['check_hostname']", bool, type(check_hostname), check_hostname
        ))

    # Don't allow 'key_file' to be provided without the 'cert_file':
    if 'key_file' in sslconfig and 'cert_file' not in sslconfig:
        raise ValueError(
            "sslconfig['key_file'] provided without sslconfig['cert_file']"
        )

    # For safety and clarity, force all paths to be absolute, normalized paths:
    for key in ('ca_file', 'ca_path', 'cert_file', 'key_file'):
        if key in sslconfig:
            value = sslconfig[key]
            if value != os.path.abspath(value):
                raise ValueError(
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(
                        key, value
                    )
                )

    sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    sslctx.verify_mode = ssl.CERT_REQUIRED
    sslctx.set_ciphers(
        'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384'
    )
    sslctx.options |= ssl.OP_NO_COMPRESSION
    if 'ca_file' in sslconfig or 'ca_path' in sslconfig:
        sslctx.load_verify_locations(
            cafile=sslconfig.get('ca_file'),
            capath=sslconfig.get('ca_path'),
        )
    else:
        if check_hostname is not True:
            raise ValueError(
                'check_hostname must be True when using default verify paths'
            )
        sslctx.set_default_verify_paths()
    if 'cert_file' in sslconfig:
        sslctx.load_cert_chain(sslconfig['cert_file'],
            keyfile=sslconfig.get('key_file')
        )
    sslctx.check_hostname = check_hostname
    return sslctx


def _validate_client_sslctx(sslctx):
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if isinstance(sslctx, dict):
        sslctx = build_client_sslctx(sslctx)

    if not isinstance(sslctx, ssl.SSLContext):
        raise TypeError('sslctx must be an ssl.SSLContext')
    if sslctx.protocol != ssl.PROTOCOL_TLSv1_2:
        raise ValueError('sslctx.protocol must be ssl.PROTOCOL_TLSv1_2')
    if not (sslctx.options & ssl.OP_NO_COMPRESSION):
        raise ValueError('sslctx.options must include ssl.OP_NO_COMPRESSION')
    if sslctx.verify_mode != ssl.CERT_REQUIRED:
        raise ValueError('sslctx.verify_mode must be ssl.CERT_REQUIRED')
    return sslctx


def _build_host(default_port, host, port, *extra):
    """
    Build value for HTTP "host" header.

    For example, for a DNS *host* name:

    >>> _build_host(80, 'en.wikipedia.org', 80)
    'en.wikipedia.org'
    >>> _build_host(80, 'en.wikipedia.org', 1234)
    'en.wikipedia.org:1234'

    And for an IPv4 literal *host*:

    >>> _build_host(80, '208.80.154.224', 80)
    '208.80.154.224'
    >>> _build_host(80, '208.80.154.224', 1234)
    '208.80.154.224:1234'

    And for an IPv6 literal *host*:

    >>> _build_host(80, '2620:0:861:ed1a::1', 80, 0, 0)
    '[2620:0:861:ed1a::1]'
    >>> _build_host(80, '2620:0:861:ed1a::1', 1234, 0, 0)
    '[2620:0:861:ed1a::1]:1234'

    """
    if not isinstance(default_port, int):
        raise TypeError(
            _TYPE_ERROR.format('default_port', int, type(default_port), default_port)
        )
    if not isinstance(host, str):
        raise TypeError(
            _TYPE_ERROR.format('host', str, type(host), host)
        )
    if not isinstance(port, int):
        raise TypeError(
            _TYPE_ERROR.format('port', int, type(port), port)
        )
    for arg in extra:
        assert isinstance(arg, int)
    if ':' in host:
        host = '[{}]'.format(host)
    if port == default_port:
        return host
    return '{}:{}'.format(host, port)


class Client:
    """
    Specifies where an HTTP server is, and how to connect to it.

    >>> client = Client(('en.wikipedia.org', 80))

    A Client is stateless and thread-safe, does not itself reference any socket
    resources.

    To make HTTP requests, create a Connection using Client.connect().
    """

    _default_port = 80  # Needed to construct the default host header
    _options = ('host', 'timeout', 'on_connect')
    __slots__ = ('address', 'options', '_family', '_base_headers') + _options

    def __init__(self, address, **options):
        # address:
        if isinstance(address, tuple):  
            if len(address) == 4:
                self._family = socket.AF_INET6
            elif len(address) == 2:
                self._family = None
            else:
                raise ValueError(
                    'address: must have 2 or 4 items; got {!r}'.format(address)
                )
            host = _build_host(self.__class__._default_port, *address)
        elif isinstance(address, (str, bytes)):
            self._family = socket.AF_UNIX
            host = None
            if isinstance(address, str) and os.path.abspath(address) != address:
                raise ValueError(
                    'address: bad socket filename: {!r}'.format(address)
                )
        else:
            raise TypeError(
                _TYPE_ERROR.format('address', (tuple, str, bytes), type(address), address)
            )
        self.address = address

        # options:
        if not set(options).issubset(self.__class__._options):
            unsupported = sorted(set(options) - set(self.__class__._options))
            raise TypeError(
                'unsupported {}() **options: {}'.format(
                    self.__class__.__name__, ', '.join(unsupported)
                )
            )
        self.options = options
        self.host = options.get('host', host)
        self.timeout = options.get('timeout', 65)
        self.on_connect = options.get('on_connect')
        assert self.host is None or isinstance(self.host, str)
        assert self.timeout is None or isinstance(self.timeout, (int, float))
        if not (self.on_connect is None or callable(self.on_connect)):
            raise TypeError(
                'on_connect: not callable: {!r}'.format(self.on_connect)
            )

        # Build _base_headers:
        self._base_headers = ({'host': self.host} if self.host else None)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.address)

    def create_socket(self):
        if self._family is None:
            return socket.create_connection(self.address, timeout=self.timeout)
        sock = socket.socket(self._family, socket.SOCK_STREAM)
        if self._family == socket.AF_UNIX:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_PASSCRED, 1)
        sock.settimeout(self.timeout)
        sock.connect(self.address)
        return sock

    def connect(self):
        sock = self.create_socket()
        conn = Connection(sock, self._base_headers)
        if self.on_connect is None or self.on_connect(conn) is True:
            return conn
        conn.close()
        raise ValueError('on_connect() did not return True')


class SSLClient(Client):
    """
    Specifies where an HTTPS server is, and how to connect to it.

    >>> sslclient = SSLClient({}, ('www.wikipedia.org', 443))

    An SSLClient is stateless and thread-safe, does not itself reference any
    socket resources.

    To make HTTP requests, create a Connection using Client.connect().
    """

    _default_port = 443  # Needed to construct the default host header
    _options = Client._options + ('ssl_host',)
    __slots__ = ('sslctx', 'ssl_host')

    def __init__(self, sslctx, address, **options):
        self.sslctx = _validate_client_sslctx(sslctx)
        super().__init__(address, **options)
        ssl_host = (address[0] if isinstance(address, tuple) else None)
        self.ssl_host = options.get('ssl_host', ssl_host)

    def __repr__(self):
        return '{}(<sslctx>, {!r})'.format(
            self.__class__.__name__, self.address
        )

    def create_socket(self):
        sock = super().create_socket()
        return self.sslctx.wrap_socket(sock, server_hostname=self.ssl_host)

